/*
 * linux/fs/bfs/file.c
 *
 * Copyright (C) 1999 Makoto Kato (m_kato@ga2.so-net.ne.jp)
 */

#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/locks.h>
#include <linux/mm.h>
#include <linux/pagemap.h>

#define	NBUF	32

#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

ssize_t bfs_file_read (struct file *, char *,  size_t, loff_t *);


static struct file_operations bfs_file_ops =
{
	NULL,				/* lseek */
	bfs_file_read,			/* read */
	NULL,				/* write */
	NULL,				/* readdir - bad */
	NULL,				/* poll - default */
	NULL,				/* ioctl - default */
	generic_file_mmap,		/* mmap */
	NULL,
	NULL,				/* flush */
	NULL,				/* release */
	file_fsync,			/* fsync */
};

struct inode_operations bfs_file_inode_operations =
{
	&bfs_file_ops,			/* default file operations */
	NULL,				/* create */
	NULL,				/* lookup */
	NULL,				/* link */
	NULL,				/* unlink */
	NULL,				/* symlink */
	NULL,				/* mkdir */
	NULL,				/* rmdir */
	NULL,				/* mknod */
	NULL,				/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	generic_readpage,		/* readpage */
	NULL,				/* writepage */
	&bfs_bmap,			/* bmap */
	NULL,				/* truncate */
	NULL,				/* permission */
	NULL				/* smap */
};


ssize_t bfs_file_read(struct file *filp, char *buf,  size_t count, loff_t *ppos)
{
	struct inode *       inode = filp->f_dentry->d_inode;
	struct super_block * sb = inode->i_sb;
	bfs_data_stream *    ds = &(inode->u.bfs_i.data);
	struct buffer_head * bh;
	bfs_inode_addr       iaddr;
	bfs_inode_addr       iaddr2;
	off_t                pos = *ppos;
	off_t                sum = 0;
	int                  i;
	int                  j;
	off_t                read_count;
	off_t                offset;

	BFS_OUTPUT (("---> Enter bfs_file_read inode %08x count %x ppos %x\n",
		inode->i_ino, count, *ppos));

	/* Get start position */
	for (i = 0; ;) {
		iaddr = bfs_read_data_stream (sb, ds, &i);
		if (iaddr.len == 0) {
			BFS_OUTPUT ((" end data_stream\n"));
			break;
		}

		if ((pos >= sum)
			&& (pos < sum + iaddr.len * sb->u.bfs_sb.block_size)) {
			iaddr.start += (pos - sum) / sb->u.bfs_sb.block_size;
			iaddr.len -= (pos - sum) / sb->u.bfs_sb.block_size;
			break;
		}
		sum += iaddr.len * sb->u.bfs_sb.block_size;
		
	}

	read_count = 0;
	offset = (pos - sum) % sb->u.bfs_sb.block_size;
	iaddr2 = iaddr;

	count = inode->i_size > pos + count ? count : inode->i_size - pos;

	for ( j = 0; j < iaddr.len; j++)
	{
		bh = bfs_bread2 (sb, iaddr2);
		if (!bh)
			return -EIO;

		BFS_OUTPUT ((" read_count %Ld offset %Ld\n", read_count, offset));

		if (count > sb->u.bfs_sb.block_size) {
			memcpy (buf + read_count, bh->b_data + offset, sb->u.bfs_sb.block_size - offset);

			*ppos += sb->u.bfs_sb.block_size - offset;
			count -= sb->u.bfs_sb.block_size - offset;
			read_count += sb->u.bfs_sb.block_size - offset;
			offset = 0;

			iaddr2.start += 1;

			brelse (bh);
		} else {
			memcpy (buf + read_count, bh->b_data + offset, count);

			*ppos += count;
			read_count += count;
	
			brelse (bh);
			break;
		}
	}

	ppos += read_count;

	BFS_OUTPUT (("  return value %d, ppos %Ld\n", read_count, *ppos));

	return read_count;
}



bfs_inode_addr bfs_read_data_stream (struct super_block * sb,
	bfs_data_stream * ds, int * pos)
{
        bfs_inode_addr       iaddr;
	dev_t                dev = sb->s_dev;
	struct buffer_head * bh;

	BFS_OUTPUT (("---> Enter bfs_read_data_stream "));
	BFS_OUTPUT (("pos %d\n", *pos));

	iaddr.allocation_group = 0;
	iaddr.start = 0;
	iaddr.len = 0;

        if( *pos < 0 )
                return iaddr;

        if (*pos < BFS_NUM_DIRECT_BLOCKS) {
		BFS_OUTPUT ((" read in direct block\n"));

                if (ds->direct[*pos].allocation_group == 0
                        && ds->direct[*pos].start == 0
			&& ds->direct[*pos].len == 0) {
		} else {
			iaddr = ds->direct[(*pos)++];
		}
        } else if (*pos < (sb->s_blocksize / sizeof(bfs_inode_addr)
		+ BFS_NUM_DIRECT_BLOCKS)) {
		bfs_inode_addr * array;

		BFS_OUTPUT ((" read in in-direct block\n"));

		bh = bfs_bread2 (sb, ds->indirect);
		if (!bh) {
			printk ("BFS: cannot read indirect block\n");
			return iaddr;
		}

		array = (bfs_inode_addr *) bh->b_data;
		iaddr = array[*pos - BFS_NUM_DIRECT_BLOCKS];

		brelse(bh);
		(*pos)++;
        } else {
		BFS_OUTPUT ((" read in double-in-direct block\n"));
        }

	if (iaddr.allocation_group == 0 && iaddr.start == 0 && iaddr.len == 0)
		*pos = 0;

        return iaddr;
}
