/*
 * linux/fs/bfs/super.c
 *
 * Copyright (C) 1999  Makoto Kato (m_kato@ga2.so-net.ne.jp)
 */

#include <linux/module.h>

#include <stdarg.h>

#include <asm/bitops.h>
#include <asm/uaccess.h>
#include <asm/system.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/bfs_fs.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/locks.h>
#include <linux/blkdev.h>
#include <linux/init.h>


void bfs_error (struct super_block * sb, const char * function, const char * fmt, ...)
{
}


NORET_TYPE void bfs_panic (struct super_block * sb, const char * function, const char * fmt, ...)
{
}


void bfs_warning (struct super_block * sb, const char * function, const char * fmt, ...)
{
}


void bfs_put_super (struct super_block * sb)
{
	MOD_DEC_USE_COUNT;
	return;
}


static const struct super_operations bfs_sops =
{
	bfs_read_inode,			/* read_inode */
	NULL,				/* write_inode */
	NULL,				/* put_inode */
	NULL,				/* delete_inode */
	NULL,				/* notify_change */
	bfs_put_super,			/* put_super */
	NULL,				/* write_super */
	bfs_statfs,			/* statfs */
	bfs_remount			/* remount_fs */
};


static int parse_options (char *options, unsigned long *sb_block,
	unsigned short *resuid, unsigned short * resgid,
	unsigned long * mount_options)
{
	return 0;
}



struct super_block * bfs_read_super (struct super_block *sb, void *data,
	int silent )
{
	bfs_super_block *    bs;
	kdev_t               dev = sb->s_dev;
	int                  blocksize = BLOCK_SIZE;
	unsigned long        logic_sb_block = 1;
	struct buffer_head * bh;

	BFS_OUTPUT (("--->Enter bfs_read_super\n"));

	/*
	 * Get current blocksize.
	 */
	blocksize = get_hardblocksize(dev);

	/* Set dummy block size */
	blocksize = 512;

	MOD_INC_USE_COUNT;
	lock_super (sb);
	set_blocksize (dev, blocksize);

	/*
	 * Get BFS super block.
	 */
	if (!(bh = bread (dev, logic_sb_block, blocksize))) {
		printk ("BFS: unable to read superblock\n");
		goto bad_bfs_read_super;
	}

	bs = (bfs_super_block *) bh->b_data;

	BFS_DUMP_SUPER_BLOCK (bs);

	/* Check magic headers of super block */
	if ((bs->magic1 != BFS_SUPER_BLOCK_MAGIC1)
	  || (bs->magic2 != BFS_SUPER_BLOCK_MAGIC2)
	  || (bs->magic3 != BFS_SUPER_BLOCK_MAGIC3)) {
		printk("BFS: different magic header\n");
		goto bad_bfs_read_super;
	}

	/*
	 * Check blocksize of BFS.
	 *
	 * Blocksize of BFS is 1024, 2048, 4096 or 8192.
	 */
	if( (bs->block_size != 1024)
	  && (bs->block_size != 2048)
	  && (bs->block_size != 4096)
	  && (bs->block_size != 8192)) {
		printk("BFS: different blocksize\n");
		goto bad_bfs_read_super;
	}

	set_blocksize (dev, (int) bs->block_size);

	/*
	 * set fs read only
	 */
	sb->s_flags |= MS_RDONLY;

	/*
	 * fill in standard stuff
	 */
	sb->s_magic = BFS_SUPER_MAGIC;
	sb->s_blocksize = (int) bs->block_size;

	sb->u.bfs_sb.block_size = bs->block_size;
	sb->u.bfs_sb.block_shift = bs->block_shift;
	sb->u.bfs_sb.num_blocks = bs->num_blocks;
	sb->u.bfs_sb.used_blocks = bs->used_blocks;
	sb->u.bfs_sb.inode_size = bs->inode_size;

	sb->u.bfs_sb.blocks_per_ag = bs->blocks_per_ag;
	sb->u.bfs_sb.ag_shift = bs->ag_shift;
	sb->u.bfs_sb.num_ags = bs->num_ags;

	sb->u.bfs_sb.log_blocks = bs->log_blocks;
	sb->u.bfs_sb.log_start = bs->log_start;
	sb->u.bfs_sb.log_end = bs->log_end;

	sb->u.bfs_sb.root_dir = bs->root_dir;
	sb->u.bfs_sb.indices = bs->indices;

	unlock_super (sb);

	/*
	 * set up enough so that it can read an inode
	 */
	sb->s_dev = dev;
	sb->s_op = (struct super_operations *) &bfs_sops;
	sb->s_root = d_alloc_root(iget(sb,
		BFSI2INODE(&(bs->root_dir),bs)), NULL);

	if (!sb->s_root) {
		sb->s_dev = 0;
		brelse (bh);
		printk ("BFS: get root inode failed\n");

		goto uninit_last_bfs_read_super;
	}

	return sb;

bad_bfs_read_super:
	sb->s_dev = 0;
	unlock_super (sb);

uninit_last_bfs_read_super:
	MOD_DEC_USE_COUNT;

	return NULL;
}


int bfs_remount (struct super_block * sb, int * flags, char * data)
{	
	if (!(*flags & MS_RDONLY))
		return -EINVAL;
	return 0;
}


static struct file_system_type bfs_fs_type = {
	"bfs", 
	FS_REQUIRES_DEV /* | FS_IBASKET */,	/* ibaskets have unresolved bugs */
	bfs_read_super, 
	NULL
};


__initfunc(int init_bfs_fs(void))
{
	return register_filesystem(&bfs_fs_type);
}

#ifdef MODULE
EXPORT_NO_SYMBOLS;

int init_module(void)
{
	return init_bfs_fs();
}


void cleanup_module(void)
{
	unregister_filesystem(&bfs_fs_type);
}

#endif


int bfs_statfs (struct super_block * sb, struct statfs * buf, int bufsiz)
{
	struct statfs tmp;

	BFS_OUTPUT (("---> Enter bfs_statfs\n"));

	tmp.f_type = BFS_SUPER_MAGIC;
	tmp.f_bsize = sb->s_blocksize;
	tmp.f_blocks = sb->u.bfs_sb.num_blocks;
	tmp.f_bfree = sb->u.bfs_sb.num_blocks - sb->u.bfs_sb.used_blocks;
	tmp.f_bavail = sb->u.bfs_sb.used_blocks;
	tmp.f_files = 0;
	tmp.f_ffree = 0;
	tmp.f_namelen = BFS_NAME_LEN;
	
	return copy_to_user(buf, &tmp, bufsiz) ? -EFAULT : 0;
}
